{% extends "base.html" %}
{% block title %}Success!{% endblock title %}

{% block content %}
  <h2>Your sketch was created with success!!!</h2>
  <p>Now you can start to draw, I mean, to code!</p>
  <p>Edit your Python sketch file located at {{ files.sketch_py.resolve() }} and have fun!</p>
  <p><a href="{{ sketch_url }}">Click here</a> to see your results or <a href="/">go back to the home</a>.</p>
{% endblock content %}
