<!DOCTYPE html>
<html>

<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>pyp5js - {% block title %}{% endblock title %}</title>

    <link href="/static/styles/basscss.min.css" rel="stylesheet">
    <link href="/static/styles/custom.css" rel="stylesheet">

  {% block extra_head %}
  {% endblock extra_head %}
</head>

<body>
  <div class="mx-auto px2 lg-px4">
  {% block content %}
  {% endblock content %}
  </div>


  {% block custom_js %}
  {% endblock custom_js %}
</body>

</html>
