{% extends "base.html" %}
{% block title %}Index{% endblock title %}

{% block content %}
  <h2>Welcome to your pyp5js sketchbook!</h2>
  <p><a href="/new-sketch/">Add a new sketch</a></p>
  <p>Available sketches in {{ sketches_dir }}:</p>
  <ul>
  {% for sketch in sketches %}
    <li><a href="{{ sketch.url }}">{{ sketch.name }}</a></li>
  {% endfor %}
  </ul>
{% endblock content %}
