<!DOCTYPE html>

<!-- pyp5js index.html boilerplate -->
<html lang="">
  <head>

    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ sketch_name }} - pyp5js</title>
    <style> body, html, canvas {padding: 0; margin: 0; overflow: hidden;} </style>

    <script src="{{ p5_js_url }}"></script>
    <script src="{{ sketch_js_url }}"  type="module"></script>
  </head>

  <body>
    <div id="sketch-holder">
          <!-- You sketch will go here! -->
    </div>
  </body>
</html>
