{% extends "base.html" %}
{% block title %}New sketch{% endblock title %}

{% block content %}
<div>
  <h2 class="h2">Start a new sketch</h2>
  <p class="regular">
    Your sketch will be created in <b>{{ sketches_dir }}</b>
  </p>
  <form action="." method="POST">
      {% if error %}
      <p class="regular red">{{error}}</p>
      {% endif %}
      <div class="flex flex-wrap items-center clearfix">
          <label class="label mr1 mb1" for="sketch_name">Sketch name</label>
          <input class="input rounded border p1 mr1 mb1 sm-col-3" type="text" name="sketch_name">
          <input class="btn rounded border-none py1 px2 mb1" type="submit" value="Create">
      </div>
  </form>
</div>
{% endblock content %}
